@@ -1,9 +1,8 @@
 import pytest
 
-from pathlib import Path
 import threading
 import socket as stdlib_socket
-import ssl as stdlib_ssl
+import ssl
 from contextlib import contextmanager
 from functools import partial
 
@@ -14,11 +13,11 @@ from async_generator import async_generator, yield_, asynccontextmanager
 import trio
 from .. import _core
 from .._highlevel_socket import SocketStream, SocketListener
-from .._highlevel_generic import BrokenStreamError, aclose_forcefully
-from .._core import ClosedResourceError
+from .._highlevel_generic import aclose_forcefully
+from .._core import ClosedResourceError, BrokenResourceError
 from .._highlevel_open_tcp_stream import open_tcp_stream
-from .. import ssl as tssl
 from .. import socket as tsocket
+from .._ssl import SSLStream, SSLListener, NeedHandshakeError
 from .._util import ConflictDetector
 
 from .._core.tests.tutil import slow
@@ -52,12 +51,20 @@ from ..testing import (
 TRIO_TEST_CA = trustme.CA()
 TRIO_TEST_1_CERT = TRIO_TEST_CA.issue_server_cert("trio-test-1.example.org")
 
-SERVER_CTX = stdlib_ssl.create_default_context(stdlib_ssl.Purpose.CLIENT_AUTH)
+SERVER_CTX = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
 TRIO_TEST_1_CERT.configure_cert(SERVER_CTX)
 
-CLIENT_CTX = stdlib_ssl.create_default_context()
+CLIENT_CTX = ssl.create_default_context()
 TRIO_TEST_CA.configure_trust(CLIENT_CTX)
 
+# Temporarily disable TLSv1.3, until the issue with openssl's session
+# ticket handling is sorted out one way or another:
+#     https://github.com/python-trio/trio/issues/819
+#     https://github.com/openssl/openssl/issues/7948
+#     https://github.com/openssl/openssl/issues/7967
+if hasattr(ssl, "OP_NO_TLSv1_3"):
+    CLIENT_CTX.options |= ssl.OP_NO_TLSv1_3
+
 
 # The blocking socket server.
 def ssl_echo_serve_sync(sock, *, expect_fail=False):
@@ -72,11 +79,11 @@ def ssl_echo_serve_sync(sock, *, expect_fail=False):
                 # other side has initiated a graceful shutdown; we try to
                 # respond in kind but it's legal for them to have already gone
                 # away.
-                exceptions = (BrokenPipeError,)
+                exceptions = (BrokenPipeError, ssl.SSLZeroReturnError)
                 # Under unclear conditions, CPython sometimes raises
                 # SSLWantWriteError here. This is a bug (bpo-32219), but it's
                 # not our bug, so ignore it.
-                exceptions += (stdlib_ssl.SSLWantWriteError,)
+                exceptions += (ssl.SSLWantWriteError,)
                 try:
                     wrapped.unwrap()
                 except exceptions:
@@ -87,6 +94,7 @@ def ssl_echo_serve_sync(sock, *, expect_fail=False):
         if expect_fail:
             print("ssl_echo_serve_sync got error as expected:", exc)
         else:  # pragma: no cover
+            print("ssl_echo_serve_sync got unexpected error:", exc)
             raise
     else:
         if expect_fail:  # pragma: no cover
@@ -120,7 +128,7 @@ async def ssl_echo_server_raw(**kwargs):
 async def ssl_echo_server(**kwargs):
     async with ssl_echo_server_raw(**kwargs) as sock:
         await yield_(
-            tssl.SSLStream(
+            SSLStream(
                 sock, CLIENT_CTX, server_hostname="trio-test-1.example.org"
             )
         )
@@ -143,7 +151,7 @@ class PyOpenSSLEchoStream:
         # TLSv1_2_METHOD.
         #
         # Discussion: https://github.com/pyca/pyopenssl/issues/624
-        if hasattr(SSL, "OP_NO_TLSv1_3"):  # pragma: no cover
+        if hasattr(SSL, "OP_NO_TLSv1_3"):
             ctx.set_options(SSL.OP_NO_TLSv1_3)
         # Unfortunately there's currently no way to say "use 1.3 or worse", we
         # can only disable specific versions. And if the two sides start
@@ -273,28 +281,28 @@ async def test_PyOpenSSLEchoStream_gives_resource_busy_errors():
     # PyOpenSSLEchoStream will notice and complain.
 
     s = PyOpenSSLEchoStream()
-    with pytest.raises(_core.ResourceBusyError) as excinfo:
+    with pytest.raises(_core.BusyResourceError) as excinfo:
         async with _core.open_nursery() as nursery:
             nursery.start_soon(s.send_all, b"x")
             nursery.start_soon(s.send_all, b"x")
     assert "simultaneous" in str(excinfo.value)
 
     s = PyOpenSSLEchoStream()
-    with pytest.raises(_core.ResourceBusyError) as excinfo:
+    with pytest.raises(_core.BusyResourceError) as excinfo:
         async with _core.open_nursery() as nursery:
             nursery.start_soon(s.send_all, b"x")
             nursery.start_soon(s.wait_send_all_might_not_block)
     assert "simultaneous" in str(excinfo.value)
 
     s = PyOpenSSLEchoStream()
-    with pytest.raises(_core.ResourceBusyError) as excinfo:
+    with pytest.raises(_core.BusyResourceError) as excinfo:
         async with _core.open_nursery() as nursery:
             nursery.start_soon(s.wait_send_all_might_not_block)
             nursery.start_soon(s.wait_send_all_might_not_block)
     assert "simultaneous" in str(excinfo.value)
 
     s = PyOpenSSLEchoStream()
-    with pytest.raises(_core.ResourceBusyError) as excinfo:
+    with pytest.raises(_core.BusyResourceError) as excinfo:
         async with _core.open_nursery() as nursery:
             nursery.start_soon(s.receive_some, 1)
             nursery.start_soon(s.receive_some, 1)
@@ -304,7 +312,7 @@ async def test_PyOpenSSLEchoStream_gives_resource_busy_errors():
 @contextmanager
 def virtual_ssl_echo_server(**kwargs):
     fakesock = PyOpenSSLEchoStream(**kwargs)
-    yield tssl.SSLStream(
+    yield SSLStream(
         fakesock, CLIENT_CTX, server_hostname="trio-test-1.example.org"
     )
 
@@ -312,13 +320,13 @@ def virtual_ssl_echo_server(**kwargs):
 def ssl_wrap_pair(
     client_transport, server_transport, *, client_kwargs={}, server_kwargs={}
 ):
-    client_ssl = tssl.SSLStream(
+    client_ssl = SSLStream(
         client_transport,
         CLIENT_CTX,
         server_hostname="trio-test-1.example.org",
         **client_kwargs
     )
-    server_ssl = tssl.SSLStream(
+    server_ssl = SSLStream(
         server_transport, SERVER_CTX, server_side=True, **server_kwargs
     )
     return client_ssl, server_ssl
@@ -334,18 +342,6 @@ def ssl_lockstep_stream_pair(**kwargs):
     return ssl_wrap_pair(client_transport, server_transport, **kwargs)
 
 
-def test_exports():
-    # Just a quick check to make sure _reexport isn't totally broken
-    assert hasattr(tssl, "SSLError")
-    assert "SSLError" in tssl.__all__
-
-    assert hasattr(tssl, "Purpose")
-    assert "Purpose" in tssl.__all__
-
-    # Intentionally omitted
-    assert not hasattr(tssl, "SSLContext")
-
-
 # Simple smoke test for handshake/send/receive/shutdown talking to a
 # synchronous server, plus make sure that we do the bare minimum of
 # certificate checking (even though this is really Python's responsibility)
@@ -359,31 +355,31 @@ async def test_ssl_client_basics():
 
     # Didn't configure the CA file, should fail
     async with ssl_echo_server_raw(expect_fail=True) as sock:
-        client_ctx = stdlib_ssl.create_default_context()
-        s = tssl.SSLStream(
+        client_ctx = ssl.create_default_context()
+        s = SSLStream(
             sock, client_ctx, server_hostname="trio-test-1.example.org"
         )
         assert not s.server_side
-        with pytest.raises(BrokenStreamError) as excinfo:
+        with pytest.raises(BrokenResourceError) as excinfo:
             await s.send_all(b"x")
-        assert isinstance(excinfo.value.__cause__, tssl.SSLError)
+        assert isinstance(excinfo.value.__cause__, ssl.SSLError)
 
     # Trusted CA, but wrong host name
     async with ssl_echo_server_raw(expect_fail=True) as sock:
-        s = tssl.SSLStream(
+        s = SSLStream(
             sock, CLIENT_CTX, server_hostname="trio-test-2.example.org"
         )
         assert not s.server_side
-        with pytest.raises(BrokenStreamError) as excinfo:
+        with pytest.raises(BrokenResourceError) as excinfo:
             await s.send_all(b"x")
-        assert isinstance(excinfo.value.__cause__, tssl.CertificateError)
+        assert isinstance(excinfo.value.__cause__, ssl.CertificateError)
 
 
 async def test_ssl_server_basics():
     a, b = stdlib_socket.socketpair()
     with a, b:
         server_sock = tsocket.from_stdlib_socket(b)
-        server_transport = tssl.SSLStream(
+        server_transport = SSLStream(
             SocketStream(server_sock), SERVER_CTX, server_side=True
         )
         assert server_transport.server_side
@@ -412,8 +408,8 @@ async def test_ssl_server_basics():
 async def test_attributes():
     async with ssl_echo_server_raw(expect_fail=True) as sock:
         good_ctx = CLIENT_CTX
-        bad_ctx = stdlib_ssl.create_default_context()
-        s = tssl.SSLStream(
+        bad_ctx = ssl.create_default_context()
+        s = SSLStream(
             sock, good_ctx, server_hostname="trio-test-1.example.org"
         )
 
@@ -421,7 +417,7 @@ async def test_attributes():
 
         # Forwarded attribute getting
         assert s.context is good_ctx
-        assert s.server_side == False
+        assert s.server_side == False  # noqa
         assert s.server_hostname == "trio-test-1.example.org"
         with pytest.raises(AttributeError):
             s.asfdasdfsa
@@ -443,9 +439,9 @@ async def test_attributes():
         # fails:
         s.context = bad_ctx
         assert s.context is bad_ctx
-        with pytest.raises(BrokenStreamError) as excinfo:
+        with pytest.raises(BrokenResourceError) as excinfo:
             await s.do_handshake()
-        assert isinstance(excinfo.value.__cause__, tssl.SSLError)
+        assert isinstance(excinfo.value.__cause__, ssl.SSLError)
 
 
 # Note: this test fails horribly if we force TLS 1.2 and trigger a
@@ -665,28 +661,28 @@ async def test_resource_busy_errors():
             await s.wait_send_all_might_not_block()
 
     s, _ = ssl_lockstep_stream_pair()
-    with pytest.raises(_core.ResourceBusyError) as excinfo:
+    with pytest.raises(_core.BusyResourceError) as excinfo:
         async with _core.open_nursery() as nursery:
             nursery.start_soon(do_send_all)
             nursery.start_soon(do_send_all)
     assert "another task" in str(excinfo.value)
 
     s, _ = ssl_lockstep_stream_pair()
-    with pytest.raises(_core.ResourceBusyError) as excinfo:
+    with pytest.raises(_core.BusyResourceError) as excinfo:
         async with _core.open_nursery() as nursery:
             nursery.start_soon(do_receive_some)
             nursery.start_soon(do_receive_some)
     assert "another task" in str(excinfo.value)
 
     s, _ = ssl_lockstep_stream_pair()
-    with pytest.raises(_core.ResourceBusyError) as excinfo:
+    with pytest.raises(_core.BusyResourceError) as excinfo:
         async with _core.open_nursery() as nursery:
             nursery.start_soon(do_send_all)
             nursery.start_soon(do_wait_send_all_might_not_block)
     assert "another task" in str(excinfo.value)
 
     s, _ = ssl_lockstep_stream_pair()
-    with pytest.raises(_core.ResourceBusyError) as excinfo:
+    with pytest.raises(_core.BusyResourceError) as excinfo:
         async with _core.open_nursery() as nursery:
             nursery.start_soon(do_wait_send_all_might_not_block)
             nursery.start_soon(do_wait_send_all_might_not_block)
@@ -700,8 +696,8 @@ async def test_wait_writable_calls_underlying_wait_writable():
         async def wait_send_all_might_not_block(self):
             record.append("ok")
 
-    ctx = stdlib_ssl.create_default_context()
-    s = tssl.SSLStream(NotAStream(), ctx, server_hostname="x")
+    ctx = ssl.create_default_context()
+    s = SSLStream(NotAStream(), ctx, server_hostname="x")
     await s.wait_send_all_might_not_block()
     assert record == ["ok"]
 
@@ -910,7 +906,7 @@ async def test_send_all_fails_in_the_middle():
     with pytest.raises(KeyError):
         await client.send_all(b"x")
 
-    with pytest.raises(BrokenStreamError):
+    with pytest.raises(BrokenResourceError):
         await client.wait_send_all_might_not_block()
 
     closed = 0
@@ -929,15 +925,15 @@ async def test_send_all_fails_in_the_middle():
 async def test_ssl_over_ssl():
     client_0, server_0 = memory_stream_pair()
 
-    client_1 = tssl.SSLStream(
+    client_1 = SSLStream(
         client_0, CLIENT_CTX, server_hostname="trio-test-1.example.org"
     )
-    server_1 = tssl.SSLStream(server_0, SERVER_CTX, server_side=True)
+    server_1 = SSLStream(server_0, SERVER_CTX, server_side=True)
 
-    client_2 = tssl.SSLStream(
+    client_2 = SSLStream(
         client_1, CLIENT_CTX, server_hostname="trio-test-1.example.org"
     )
-    server_2 = tssl.SSLStream(server_1, SERVER_CTX, server_side=True)
+    server_2 = SSLStream(server_1, SERVER_CTX, server_side=True)
 
     async def client():
         await client_2.send_all(b"hi")
@@ -961,9 +957,9 @@ async def test_ssl_bad_shutdown():
 
     await trio.aclose_forcefully(client)
     # now the server sees a broken stream
-    with pytest.raises(BrokenStreamError):
+    with pytest.raises(BrokenResourceError):
         await server.receive_some(10)
-    with pytest.raises(BrokenStreamError):
+    with pytest.raises(BrokenResourceError):
         await server.send_all(b"x" * 10)
 
     await server.aclose()
@@ -982,7 +978,7 @@ async def test_ssl_bad_shutdown_but_its_ok():
     await trio.aclose_forcefully(client)
     # the server sees that as a clean shutdown
     assert await server.receive_some(10) == b""
-    with pytest.raises(BrokenStreamError):
+    with pytest.raises(BrokenResourceError):
         await server.send_all(b"x" * 10)
 
     await server.aclose()
@@ -995,8 +991,8 @@ async def test_ssl_handshake_failure_during_aclose():
     # the underlying transport.
     async with ssl_echo_server_raw(expect_fail=True) as sock:
         # Don't configure trust correctly
-        client_ctx = stdlib_ssl.create_default_context()
-        s = tssl.SSLStream(
+        client_ctx = ssl.create_default_context()
+        s = SSLStream(
             sock, client_ctx, server_hostname="trio-test-1.example.org"
         )
         # It's a little unclear here whether aclose should swallow the error
@@ -1005,7 +1001,7 @@ async def test_ssl_handshake_failure_during_aclose():
         # simultaneously is allowed. But I guess when https_compatible=False
         # then it's bad if we can get through a whole connection with a peer
         # that has no valid certificate, and never raise an error.
-        with pytest.raises(BrokenStreamError) as excinfo:
+        with pytest.raises(BrokenResourceError):
             await s.aclose()
 
 
@@ -1047,9 +1043,9 @@ async def test_ssl_https_compatibility_disagreement():
     # client is in HTTPS-mode, server is not
     # so client doing graceful_shutdown causes an error on server
     async def receive_and_expect_error():
-        with pytest.raises(BrokenStreamError) as excinfo:
+        with pytest.raises(BrokenResourceError) as excinfo:
             await server.receive_some(10)
-        assert isinstance(excinfo.value.__cause__, tssl.SSLEOFError)
+        assert isinstance(excinfo.value.__cause__, ssl.SSLEOFError)
 
     async with _core.open_nursery() as nursery:
         nursery.start_soon(client.aclose)
@@ -1082,7 +1078,7 @@ async def test_send_error_during_handshake():
         with assert_checkpoints():
             await client.do_handshake()
 
-    with pytest.raises(BrokenStreamError):
+    with pytest.raises(BrokenResourceError):
         with assert_checkpoints():
             await client.do_handshake()
 
@@ -1105,11 +1101,87 @@ async def test_receive_error_during_handshake():
         nursery.start_soon(client_side, nursery.cancel_scope)
         nursery.start_soon(server.do_handshake)
 
-    with pytest.raises(BrokenStreamError):
+    with pytest.raises(BrokenResourceError):
         with assert_checkpoints():
             await client.do_handshake()
 
 
+async def test_selected_alpn_protocol_before_handshake():
+    client, server = ssl_memory_stream_pair()
+
+    with pytest.raises(NeedHandshakeError):
+        client.selected_alpn_protocol()
+
+    with pytest.raises(NeedHandshakeError):
+        server.selected_alpn_protocol()
+
+
+async def test_selected_alpn_protocol_when_not_set():
+    # ALPN protocol still returns None when it's not ser,
+    # instead of raising an exception
+    client, server = ssl_memory_stream_pair()
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(client.do_handshake)
+        nursery.start_soon(server.do_handshake)
+
+    assert client.selected_alpn_protocol() is None
+    assert server.selected_alpn_protocol() is None
+
+    assert client.selected_alpn_protocol() == \
+        server.selected_alpn_protocol()
+
+
+async def test_selected_npn_protocol_before_handshake():
+    client, server = ssl_memory_stream_pair()
+
+    with pytest.raises(NeedHandshakeError):
+        client.selected_npn_protocol()
+
+    with pytest.raises(NeedHandshakeError):
+        server.selected_npn_protocol()
+
+
+async def test_selected_npn_protocol_when_not_set():
+    # NPN protocol still returns None when it's not ser,
+    # instead of raising an exception
+    client, server = ssl_memory_stream_pair()
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(client.do_handshake)
+        nursery.start_soon(server.do_handshake)
+
+    assert client.selected_npn_protocol() is None
+    assert server.selected_npn_protocol() is None
+
+    assert client.selected_npn_protocol() == \
+        server.selected_npn_protocol()
+
+
+async def test_get_channel_binding_before_handshake():
+    client, server = ssl_memory_stream_pair()
+
+    with pytest.raises(NeedHandshakeError):
+        client.get_channel_binding()
+
+    with pytest.raises(NeedHandshakeError):
+        server.get_channel_binding()
+
+
+async def test_get_channel_binding_after_handshake():
+    client, server = ssl_memory_stream_pair()
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(client.do_handshake)
+        nursery.start_soon(server.do_handshake)
+
+    assert client.get_channel_binding() is not None
+    assert server.get_channel_binding() is not None
+
+    assert client.get_channel_binding() == \
+        server.get_channel_binding()
+
+
 async def test_getpeercert():
     # Make sure we're not affected by https://bugs.python.org/issue29334
     client, server = ssl_memory_stream_pair()
@@ -1132,10 +1204,10 @@ async def test_SSLListener():
         await listen_sock.bind(("127.0.0.1", 0))
         listen_sock.listen(1)
         socket_listener = SocketListener(listen_sock)
-        ssl_listener = tssl.SSLListener(socket_listener, SERVER_CTX, **kwargs)
+        ssl_listener = SSLListener(socket_listener, SERVER_CTX, **kwargs)
 
         transport_client = await open_tcp_stream(*listen_sock.getsockname())
-        ssl_client = tssl.SSLStream(
+        ssl_client = SSLStream(
             transport_client,
             CLIENT_CTX,
             server_hostname="trio-test-1.example.org"
